(function() {
  var ZeedhiDirectives = angular.module('ZeedhiDirectives');
  var ZhKeyboardService;

  ZeedhiDirectives.directive('zhKeyboard', [
    '$rootScope',
    'ZhKeyboardService',
    'templateManager',
    'ScreenService',
    '$timeout',
    zhKeyboardDirective
  ]);

  function zhKeyboardDirective(
    $rootScope,
    _ZhKeyboardService,
    templateManager,
    ScreenService,
    $timeout
  ) {
    ZhKeyboardService = _ZhKeyboardService;
    return {
      link: function($scope, element, attrs, controller) {
        var lastWidgetActionsBoundKeys = [];
        var changeSourceActionsDeregister = $rootScope.$on('changeSourceActions', function(event, widget) {
          unbindWidgetActions(widget, lastWidgetActionsBoundKeys);
          lastWidgetActionsBoundKeys = bindWidgetActions(widget);
        });

        $scope.$on('$destroy', function() {
          changeSourceActionsDeregister();
        });

        // default keyboard shortcuts
        ZhKeyboardService.bind('esc', escapeKeyEvent);
        ZhKeyboardService.bind('enter', enterKeyEvent);
        ZhKeyboardService.bind(['tab', 'shift+tab'], tabKeyEvent);
        ZhKeyboardService.bind(['up', 'down'], upDownKeyEvent);
        ZhKeyboardService.bind(['left', 'right'], leftRightKeyEvent);
        ZhKeyboardService.bind(['pageup', 'pagedown'], pageUpDownKeyEvent);

        if (!Util.isDesktop()) {
          // on Android devices the "NEXT" and "GO" buttons of keyboard trigger an key event
          Mousetrap.bind('tab', mobileTabKeyEvent);
          Mousetrap.bind('enter', mobileEnterKeyEvent);
        }

        templateManager.projectPromise.then(function() {
          // project keyboard shortcuts
          if (!templateManager.project.shortcutKeys) {
            templateManager.project.shortcutKeys = {};
          }

          angular.forEach(templateManager.project.shortcutKeys, function(value, key) {
            templateManager.project.shortcutKeys[key] = value
              .toLowerCase()
              .replace('ctrl', 'mod')
              .replace('command', 'mod');
          });

          ZhKeyboardService.bind(
            templateManager.project.shortcutKeys.openMenu || 'mod+m',
            openMenu
          );
          ZhKeyboardService.bind(
            templateManager.project.shortcutKeys.refreshWidget || 'f5',
            refreshWidget
          );
          ZhKeyboardService.bind(
            templateManager.project.shortcutKeys.openMenuOpenedWindows || 'mod+o',
            openMenuOpenedWindows
          );
          ZhKeyboardService.bind(
            [
              templateManager.project.shortcutKeys.navigateTabLeft || 'alt+left',
              templateManager.project.shortcutKeys.navigateTabRight || 'alt+right'
            ],
            changeTabbedWidget
          );
          ZhKeyboardService.bind(
            templateManager.project.shortcutKeys.clearSearchInput || 'mod+del',
            clearSearchInput
          );
          ZhKeyboardService.bind(
            templateManager.project.shortcutKeys.openFieldAction || 'mod+space',
            openFieldAction
          );
          ZhKeyboardService.bind(
            templateManager.project.shortcutKeys.openRightActions || 'mod+enter',
            openRightActionsMenu
          );

          if (templateManager.project.notifications) {
            templateManager.project.notifications.forEach(function(notification) {
              if (notification.shortcutKey) {
                ZhKeyboardService.bind(notification.shortcutKey, function() {
                  $scope.notificationClick(notification);
                  return false;
                });
              }
            });
          }
        });

        function isMobiScrollOpen() {
          return $('.mbsc-mobiscroll').length > 0;
        }

        //Make the window back function (called when using the device hardware back button)
        //work the same way as the esc event
        if (Util.isMobile()) {
          window.__back__ = escapeKeyEvent;
        }

        function escapeKeyEvent() {
          var activeElement = $(document.activeElement);
          var handled = true;

          if ($rootScope.showAbout) {
            ScreenService.hideAbout();
          } else if (isFloatingButtonSearch(activeElement)) {
            activeElement.blur();
            closeFloatingControl(activeElement);
          } else if (ScreenService.isPopupOpen()) {
            closePopup();
          } else if (isMobiScrollOpen()) {
            $('.dwb-c div').click();
          } else if (ScreenService.isSideMenuOpen()) {
            ScreenService.toggleSideMenu();
          } else if (isInlineEditableInput(activeElement)) {
            activeElement[0].cancelEditing();
          } else if (hasVisibleLeftAction()) {
            clickLeftAction();
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function isFloatingButtonSearch(element) {
          return element.is('.zh-input-search-floating');
        }

        function closePopup() {
          if (currentPopupHasBackAction()) {
            var eventToTrigger = Util.isMobile() ? 'touchstart' : 'click';
            $rootScope._currentPopup
              .getActionsByVisibility('back')[0]
              .element.trigger(eventToTrigger);
          } else {
            ScreenService.closePopup();
          }
        }

        function currentPopupHasBackAction() {
          return (
            $rootScope._currentPopup &&
            $rootScope._currentPopup.getActionsByVisibility('back') &&
            $rootScope._currentPopup.getActionsByVisibility('back').length
          );
        }

        function isInlineEditableInput(element) {
          return element.is('.inline-edit');
        }

        function hasVisibleLeftAction() {
          return $('.zh-footer-left .previous:visible').length > 0;
        }

        function clickLeftAction() {
          var eventToTrigger = Util.isMobile() ? 'touchstart' : 'click';
          $('.zh-footer-left .previous:visible').trigger(eventToTrigger);
        }

        function openMenu(event, key) {
          // if (!ScreenService.isWidgetEditing() && !ScreenService.isPopupOpen() && !isMobiScrollOpen() && !ScreenService.isAlertShown()) {
          if (
            !ScreenService.isPopupOpen() &&
            !isMobiScrollOpen() &&
            !ScreenService.isAlertShown()
          ) {
            if (!ScreenService.isSideMenuOpen()) {
              ScreenService.toggleSideMenu();
            }
            Util.selectElement($('.zh-input-search-menu:visible'));
            handled = true;
          }

          // prevent default behaviour
          return false;
        }

        function _navigateMenu(key) {
          var menuSelected = $('.zh-main-menu li.menu-item.visible.selected');
          var menus = $('.zh-main-menu li.menu-item.visible');
          var handled = false;
          var newSelection;

          if (key == 'up' || key == 'down') {
            if (menuSelected.length === 0) {
              menus.eq(0).addClass('selected');
            } else {
              var selectedIndex = menus.index(menuSelected);
              if (key == 'up') {
                if (selectedIndex > 0) {
                  newSelection = menus.eq(selectedIndex - 1);
                } else if (selectedIndex === 0) {
                  menuSelected
                    .parents('.zh-menu-container')
                    .finish()
                    .animate({ scrollTop: 0 }, 500);
                  if (
                    templateManager.container.showSearch ||
                    templateManager.project.showMenuSearch
                  ) {
                    $('.zh-input-search-menu').focus();
                  }
                }
              } else if (key == 'down') {
                if (selectedIndex < menus.length - 1) {
                  newSelection = menus.eq(selectedIndex + 1);
                }
              }
            }

            handled = true;
          } else if (
            key == 'right' &&
            menuSelected.hasClass('has-child') &&
            menuSelected.hasClass('closed')
          ) {
            menuSelected.children('.menu-link').click();
            handled = true;
          } else if (key == 'left') {
            if (menuSelected.hasClass('has-child') && !menuSelected.hasClass('closed')) {
              menuSelected.children('.menu-link').click();
              handled = true;
            } else if (menuSelected.hasClass('closed')) {
              newSelection = menuSelected.parent().closest('li.menu-item');
              handled = true;
            }
          }

          if (newSelection && newSelection.length > 0) {
            menuSelected.removeClass('selected');
            newSelection.addClass('selected');
            if (newSelection.hasClass('disabled')) {
              _navigateMenu(key);
            } else {
              var container = newSelection.parents('.zh-menu-container');
              var newSelectionLink = newSelection.children('.menu-link');
              if (newSelectionLink.offset().top < 0) {
                container
                  .finish()
                  .animate(
                    { scrollTop: container.scrollTop() + newSelectionLink.offset().top - 10 },
                    500
                  );
              } else if (
                newSelectionLink.offset().top >
                container.innerHeight() - newSelectionLink.outerHeight()
              ) {
                container
                  .finish()
                  .animate(
                    {
                      scrollTop:
                        container.scrollTop() +
                        newSelectionLink.offset().top -
                        container.innerHeight() +
                        newSelectionLink.outerHeight() +
                        10
                    },
                    500
                  );
              }
            }
          }

          return handled;
        }

        function _navigateMessageOptions(key) {
          var alertContainer = $('.zh-container-alert').parent();
          var selectedButton = alertContainer.find('button:focus');
          if (key == 'up' || key == 'left' || key == 'shift+tab') {
            if (selectedButton.length > 0) {
              selectedButton = selectedButton.prev();
            }
            if (selectedButton.length === 0) {
              selectedButton = alertContainer.find('button').last();
            }
          } else if (key == 'down' || key == 'right' || key == 'tab') {
            if (selectedButton.length > 0) {
              selectedButton = selectedButton.next();
            }
            if (selectedButton.length === 0) {
              selectedButton = alertContainer.find('button').first();
            }
          }

          if (selectedButton.length > 0) {
            selectedButton.focus();
          }
        }

        function _navigateActionsList(key) {
          var actionListContainer = $('.popup-container').find('.zh-actions-list');
          var listItems = actionListContainer.find('li');
          var listSelected = actionListContainer.find('li.selected');

          if (listSelected.length === 0) {
            listItems.eq(0).addClass('selected');
          } else {
            var selectedIndex = listItems.index(listSelected);
            var newSelection;
            if (key == 'up') {
              if (selectedIndex > 0) {
                newSelection = listItems.eq(selectedIndex - 1);
              }
            } else if (key == 'down') {
              if (selectedIndex < listItems.length - 1) {
                newSelection = listItems.eq(selectedIndex + 1);
              }
            }
            if (newSelection) {
              listSelected.removeClass('selected');
              newSelection.addClass('selected');
              var container = newSelection.parents('.container').eq(0);
              if (newSelection.offset().top < container.offset().top) {
                container
                  .finish()
                  .animate(
                    {
                      scrollTop:
                        container.scrollTop() +
                        (newSelection.offset().top - container.offset().top) -
                        10
                    },
                    500
                  );
              } else if (
                newSelection.offset().top >
                container.offset().top + container.innerHeight() - newSelection.outerHeight()
              ) {
                container
                  .finish()
                  .animate(
                    {
                      scrollTop:
                        container.scrollTop() +
                        newSelection.offset().top -
                        container.offset().top -
                        container.innerHeight() +
                        newSelection.outerHeight() +
                        10
                    },
                    500
                  );
              }
            }
          }
        }

        function upDownKeyEvent(event, key) {
          var handled = true;
          var activeElement = $(document.activeElement);

          if (ScreenService.isSideMenuOpen()) {
            _navigateMenu(key);
          } else if (ScreenService.isAlertShown()) {
            _navigateMessageOptions(key);
          } else if (
            ScreenService.isPopupOpen() &&
            $('.popup-container .zh-actions-list').length > 0
          ) {
            _navigateActionsList(key);
          } else if (activeElement.is('.inline-edit')) {
            // inline editable grid -> navigate
            _navigateEditableGrid(key);
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function leftRightKeyEvent(event, key) {
          var handled = true;

          if (ScreenService.isSideMenuOpen()) {
            handled = _navigateMenu(key);
          } else if (ScreenService.isAlertShown()) {
            _navigateMessageOptions(key);
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function _isForm(widget) {
          var template = widget.template;
          return (
            template.indexOf('form-login.html') !== -1 ||
            template.indexOf('form-center-top.html') !== -1 ||
            template.indexOf('form-view-edit.html') !== -1 ||
            template.indexOf('form-view.html') !== -1 ||
            template.indexOf('form.html') !== -1
          );
        }

        function _getFocusedElement() {
          var element = $(document.activeElement);
          if (
            element.is('.zh-label-with-icon') ||
            element.parent().is('.zh-radio-label') ||
            element.parent().is('.zh-box-label') ||
            element.is('textarea') ||
            element.is('button') ||
            element.is('input')
          ) {
            return element;
          } else {
            return undefined;
          }
        }

        function _navigateEditableGrid(key) {
          var element = $(document.activeElement);
          var cell = element.parents('.td');
          var row = cell.parent();
          var tbody = row.parent();
          var cells = row.children();
          var rows = tbody.children('.tr');
          var cellIndex = cells.index(cell);
          var rowIndex = rows.index(row);
          var newSelection, input, verticalScroll;

          var switchFromElementToNewSelection = function() {
            if (element.is(':focus')) { 
              element.trigger('blur');
            }
            
            templateManager.updateTemplate().then(function() {
              if (newSelection[0].__scope) {
                newSelection[0].__scope.widget.inlineEditing = true;
              }
              newSelection.click();
            });
          };

          if (key == 'enter' || key == 'tab' || key == 'down') {
            if (key != 'down') {
              for (cellIndex++; cellIndex <= cells.length - 1; cellIndex++) {
                input = cells.eq(cellIndex).find('input.inline-edit');
                if (input.length > 0) {
                  newSelection = input.eq(0);
                  newSelection.click();
                  break;
                }
              }
            }

            if (!newSelection) {
              if (rowIndex === rows.length - 1) {
                verticalScroll = tbody.parent().siblings().eq(1);
                verticalScroll.scrollTop(verticalScroll.scrollTop() + row.outerHeight());
                rowIndex--;
              }
              $timeout(function() {
                for (rowIndex++; rowIndex <= rows.length - 1; rowIndex++) {
                  if (key == 'down') {
                    input = rows.eq(rowIndex).children().eq(cellIndex).find('input.inline-edit');
                  } else {
                    input = rows.eq(rowIndex).find('input.inline-edit');
                  }
                  if (input.length > 0) {
                    newSelection = input.eq(0);
                    switchFromElementToNewSelection();
                    break;
                  }
                }
              }, 0, false);
            }
          } else if (key == 'shift+tab' || key == 'up') {
            if (key != 'up') {
              for (cellIndex--; cellIndex >= 0; cellIndex--) {
                input = cells.eq(cellIndex).find('input.inline-edit');
                if (input.length > 0) {
                  newSelection = input.eq(0);
                  switchFromElementToNewSelection();
                  break;
                }
              }
            }

            if (!newSelection) {
              if (rowIndex === 0) {
                verticalScroll = tbody.parent().siblings().eq(1);
                verticalScroll.scrollTop(verticalScroll.scrollTop() - row.outerHeight());
                rowIndex++;
              }
              $timeout(function() {
                for (rowIndex--; rowIndex >= 0; rowIndex--) {
                  if (key == 'up') {
                    input = rows.eq(rowIndex).children().eq(cellIndex).find('input.inline-edit');
                  } else {
                    input = rows.eq(rowIndex).find('input.inline-edit');
                  }
                  if (input.length > 0) {
                    newSelection = input.last();
                    switchFromElementToNewSelection();
                    break;
                  }
                }
              }, 0, false);
            }
          }
        }

        function enterKeyEvent(event, key) {
          var focusedElement = _getFocusedElement();
          var widget = templateManager.container.getActiveWidget();
          var handled = true;

          if (focusedElement && focusedElement.is('textarea')) {
            return true;
          }

          // if (ScreenService.isWidgetEditing() && !ScreenService.isPopupOpen()) {
          // 	$('.widget-edit-footer .action-right').click();
          // } else
          if (ScreenService.isSideMenuOpen()) {
            // click on selected menu item
            var menuSelected = $('.zh-main-menu li.menu-item.selected');
            if (menuSelected.length == 1) {
              menuSelected.children('.menu-link').click();
            }
          } else if (ScreenService.isAboutShown()) {
            ScreenService.hideAbout();
          } else if (isMobiScrollOpen()) {
            $('.dwb-s div').click();
          } else if (ScreenService.isAlertShown()) {
            // click on the ok button of alert messages
            var alertContainer = $('.zh-container-alert').parent();
            alertContainer.find('button:focus').click();
          } else if (
            ScreenService.isPopupOpen() &&
            $('.popup-container .zh-actions-list').length > 0
          ) {
            // click on selected action
            var actionSelected = $('.popup-container .zh-actions-list li.selected');
            if (actionSelected.length == 1) {
              actionSelected.click();
            }
          } else if (focusedElement && focusedElement.is('.inline-edit')) {
            // inline editable grid -> navigate
            _navigateEditableGrid(key);
          } else if (focusedElement && !focusedElement.is('input')) {
            // click on focused element
            focusedElement.click();
          } else if (_isForm(widget)) {
            // click on default button of a form
            var defaultButton = widget.fields.filter(function(item) {
              return item.defaultButton;
            })[0];
            if (defaultButton) {
              defaultButton.click();
            }
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function tabKeyEvent(event, key) {
          var handled = true;

          if (ScreenService.isAlertShown()) {
            _navigateMessageOptions(key);
          } else if ($(document.activeElement).is('.inline-edit')) {
            // inline editable grid -> navigate
            _navigateEditableGrid(key);
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function mobileEnterKeyEvent(event, key) {
          var handled = true;

          if ($(document.activeElement).is('.inline-edit')) {
            // inline editable grid -> navigate
            _navigateEditableGrid(key);
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function mobileTabKeyEvent(event, key) {
          var handled = true;

          if ($(document.activeElement).is('.inline-edit')) {
            // inline editable grid -> navigate
            _navigateEditableGrid(key);
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function closeFloatingControl(activeElement) {
          var widget = ScreenService.isPopupOpen() ? $rootScope._currentPopup : templateManager.container.getActiveWidget();
          var floatingControl = (ScreenService.isPopupOpen() ? $('.popup-container .container:visible') : $('.container:visible'))
            .last()
            .parent()
            .find('.zh-floating-control');
          if (floatingControl.length == 1) {
            var floatingControlElem = $(activeElement).closest('.zh-floating-control');

            if (floatingControlElem[0].__scope) {
              floatingControlElem[0].__scope.widget.closeActionMenu();
            }
          }
        }

        var _refreshWidget = Util.buildDebounceMethod(
          function(widget) {
            widget.refreshWidget();
          },
          300,
          true
        );

        function refreshWidget(event, key) {
          var widget = ScreenService.isPopupOpen() ? $rootScope._currentPopup : templateManager.container.getActiveWidget();
          if (
            !ScreenService.isSideMenuOpen() &&
            !isMobiScrollOpen() &&
            !ScreenService.isAlertShown()
          ) {
            var floatingControl = (ScreenService.isPopupOpen() ? $('.popup-container .container:visible') : $('.container:visible'))
              .last()
              .parent()
              .find('.zh-floating-control');
            if (floatingControl.length == 1) {
              _refreshWidget(widget);
            }
          }

          // prevent default behaviour
          return false;
        }

        function clearSearchInput(event, key) {
          var activeElement = $(document.activeElement);
          if (
            activeElement.is('.zh-input-search-menu') ||
            activeElement.is('.zh-input-search-floating')
          ) {
            activeElement.parent().find('.clear-button').click();
          }

          // prevent default behaviour
          return false;
        }

        function changeTabbedWidget(event, key) {
          var activeElement = $(document.activeElement);
          if (activeElement.is('textarea') || activeElement.is('input')) {
            return;
          }
          if (
            !ScreenService.isSideMenuOpen() &&
            !isMobiScrollOpen() &&
            !ScreenService.isAlertShown()
          ) {
            var topMostContainer = $('.container[ng-controller="TabbedChild"]:visible')
              .last()
              .parent();
            if (topMostContainer.length > 0) {
              var tabbedControl = topMostContainer.find('.zh-tab-scroller');
              if (tabbedControl.length > 0) {
                var tabs = tabbedControl.find('.zh-tab-cell a');
                var activeTab = tabbedControl.find('.zh-tab-cell a.active');
                var activeTabIndex = tabs.index(activeTab);

                if (
                  key == (templateManager.project.shortcutKeys.navigateTabRight || 'alt+right') &&
                  activeTabIndex < tabs.length - 1
                ) {
                  tabs.eq(activeTabIndex + 1).click();
                } else if (
                  key == (templateManager.project.shortcutKeys.navigateTabLeft || 'alt+left') &&
                  activeTabIndex > 0
                ) {
                  tabs.eq(activeTabIndex - 1).click();
                }
              }
            }
          }

          // prevent default behaviour
          return false;
        }

        function _navigateGridList(key) {
          var topMostContainer = $('.container:visible').last();
          if (topMostContainer.hasClass('zh-list-container')) {
            var innerContainer = topMostContainer;

            if (key == 'pagedown') {
              if (
                innerContainer.height() == innerContainer[0].scrollHeight &&
                innerContainer.find('.infinity-load-data').length > 0
              ) {
                innerContainer.find('.infinity-load-data').click();
              } else {
                innerContainer
                  .finish()
                  .animate(
                    { scrollTop: innerContainer.scrollTop() + innerContainer.height() },
                    500
                  );
              }
            } else if (key == 'pageup') {
              innerContainer
                .finish()
                .animate({ scrollTop: innerContainer.scrollTop() - innerContainer.height() }, 500);
            }
          }
        }

        function pageUpDownKeyEvent(event, key) {
          var handled = true;
          // if (!ScreenService.isWidgetEditing() && !ScreenService.isSideMenuOpen() && !isMobiScrollOpen() && !ScreenService.isAlertShown()) {
          if (
            !ScreenService.isSideMenuOpen() &&
            !isMobiScrollOpen() &&
            !ScreenService.isAlertShown()
          ) {
            _navigateGridList(key);
          } else {
            handled = false;
          }

          // prevent default behaviour
          return !handled;
        }

        function _getDefaultRightAction() {
          var rightActionContainer = ScreenService.isPopupOpen() ? $('.popup-container .zh-footer-right:visible') : $('.zh-footer-right:visible');
          return rightActionContainer.length == 1 ? rightActionContainer.find('li').not('.more-actions') : undefined;
        }

        function openRightActionsMenu(event, key) {
          var rightActionContainer = ScreenService.isPopupOpen() ? $('.popup-container .zh-footer-right:visible') : $('.zh-footer-right:visible');
          var rightAction =
            rightActionContainer.length == 1 ? rightActionContainer.find('li').not('.more-actions') : undefined;
          // if (!ScreenService.isWidgetEditing() && !ScreenService.isPopupOpen() && !isMobiScrollOpen() && !ScreenService.isSideMenuOpen() && !ScreenService.isAlertShown()) {
          if (!ScreenService.isSideMenuOpen() && !ScreenService.isAlertShown()) {
            if (rightAction && rightAction.length > 0) {
              // click on right action if is on popup or in edit mode
              $(':focus').blur();
              rightAction.click();
            } else {
              // click on the back button if it is visible
              rightActionContainer.find('.more-actions:visible').click();
            }
          }

          // prevent default behaviour
          return false;
        }

        function openFieldAction(event, key) {
          var focusedElement = $(document.activeElement);
          if (focusedElement.is('.zh-select-multiple')) {
            focusedElement.click();
          }

          // prevent default behaviour
          return false;
        }

        function openMenuOpenedWindows(event, key) {
          // if (!ScreenService.isWidgetEditing() && !ScreenService.isPopupOpen() && !isMobiScrollOpen() && !ScreenService.isAlertShown()) {
          if (
            !ScreenService.isPopupOpen() &&
            !isMobiScrollOpen() &&
            !ScreenService.isAlertShown()
          ) {
            $rootScope.$broadcast('openHistoryMenu');
          }

          // prevent default behaviour
          return false;
        }
      }
    };
  }

  function bindWidgetActions(widget) {
    widget.prepareActions();
    var boundActionsKeys = widget.actions
      .map(function(action) {
        if (action.shortcutKey) {
          return ZhKeyboardService.bind(
            action.shortcutKey,
            function() {
              if (action.element && action.element.is(':visible')) {
                action.element.click();
                return false;
              }
            },
            'widget'
          );
        }
      })
      .filter(function(key) {
        return !!key;
      });

    if (Util.isArray(boundActionsKeys)) {
      // flattens the array;
      boundActionsKeys = [].concat.apply([], boundActionsKeys);
    }

    return boundActionsKeys || [];
  }

  function unbindWidgetActions(widget, keysToUnbind) {
    ZhKeyboardService.unbind(keysToUnbind);
  }
})();
